#!/usr/bin/perl
# Porovná dva soubory složkových stromů ve formátu Penn.
# Nenahrazuje vyhodnocovací program evalb, ale umožňuje analyzovat chyby.
# (c) 2007 Dan Zeman <zeman@ufal.mff.cuni.cz>
# Licence: GNU GPL

sub usage
{
    print STDERR ("Usage: cmp_penn.pl goldstandard.penn system.penn [system2.penn]\n");
}

use utf8;
use open ":utf8";
binmode(STDIN, ":utf8");
binmode(STDOUT, ":utf8");
binmode(STDERR, ":utf8");
use penn;



if(@ARGV<2)
{
    usage();
    exit(1);
}
$s1 = $ARGV[0];
$s2 = $ARGV[1];
open(S1, $s1) or die("Nelze číst ze souboru $s1: $!\n");
open(S2, $s2) or die("Nelze číst ze souboru $s2: $!\n");
if(@ARGV>=3)
{
    $s3 = $ARGV[2];
    open(S3, $s3) or die("Nelze číst ze souboru $s3: $!\n");
}
$n_chyb = 0;
while(1)
{
    $text1 = nacist_vetu(S1);
    $text2 = nacist_vetu(S2);
    last if($text1 eq "" || $text2 eq "");
    $strom1 = penn::decompose($text1);
    $strom2 = penn::decompose($text2);
    if($s3)
    {
        $text3 = nacist_vetu(S3);
        $strom3 = penn::decompose($text3);
        porovnat_3_stromy($strom1, $strom2, $strom3);
    }
    else
    {
        porovnat_stromy($strom1, $strom2);
    }
    # Upravit statistiku složek vzorového stromu.
    foreach my $slozka (@{$strom1->{slozky}})
    {
        my $pravidlo = "$slozka->{label} --> $slozka->{rhs}";
        $gramatika{$pravidlo}++;
    }
}
# Vypsat statistiku chyb.
print("Statistika chyb\n");
@chyby = sort {$chyby{$a} <=> $chyby{$b}} keys(%chyby);
foreach my $chyba (@chyby)
{
    my $podil = $n_chyb>0 ? $chyby{$chyba}/$n_chyb : 0;
    my $mezery;
    for(my $i = length($chyba); $i<15; $i++)
    {
        $mezery .= " ";
    }
    print("$chyba$mezery\t$chyby{$chyba}\t$podil\n");
}
# Vypsat gramatiku.
@klice = sort {$gramatika{$a} <=> $gramatika{$b}} keys(%gramatika);
foreach my $pravidlo (@klice)
{
    print("$pravidlo\t$gramatika{$pravidlo}\n");
}



#------------------------------------------------------------------------------
# Načte větu z otevřeného souboru. Počítá s tím, že jedna věta může být na více
# než jednom řádku. Předpokládá ale, že na jednom řádku nejsou části více než
# jedné věty.
#------------------------------------------------------------------------------
sub nacist_vetu
{
    my $SOUBOR = shift;
    # Vždy, když se vyrovná počet levých a pravých závorek, končí věta.
    # Řetězce přilepené k levým závorkám jsou neterminály.
    # Ostatní řetězce jsou vesměs slova (tokeny), ale "*-1", "*-2"..., "*U*", "*" je stopa nebo něco takového.
    my $text;
    my $nlz_celkem = 0;
    my $npz_celkem = 0;
    while(<$SOUBOR>)
    {
        s/\r?\n$//;
        # Spočítat levé a pravé závorky v právě načteném řádku.
        my ($nlz, $npz);
        my $lz = $_;
        $lz =~ s/[^(]//g;
        $nlz = length($lz);
        my $pz = $_;
        $pz =~ s/[^)]//g;
        $npz = length($pz);
        # Přičíst počty závorek k celkovým počtům a zjistit, zda jde o konec věty.
        $nlz_celkem += $nlz;
        $npz_celkem += $npz;
        my $konec_vety = $nlz_celkem>0 && $nlz_celkem==$npz_celkem;
        # Připojit aktuální řádek k větě, a pokud je věta celá, přerušit čtení a zpracovat ji.
        $text .= $_;
        last if($konec_vety);
    }
    # Jestliže text neobsahuje ani jednu závorku, nenačetli jsme větu (strom), ale nanejvýš prázdné řádky mezi.
    # Možná jsme také na konci souboru. Usnadníme detekci takových případů a text předáme rovnou zcela prázdný.
    if($text !~ m/[\(\)]/)
    {
        $text = "";
    }
    return $text;
}



#------------------------------------------------------------------------------
# Porovná složky ve dvou stromech téže věty.
#------------------------------------------------------------------------------
sub porovnat_stromy
{
    # Strom je hash s odkazy na @slova a @slozky.
    my $strom1 = shift;
    my $strom2 = shift;
    # Nahashovat složky kvůli snadnějšímu hledání.
    my %hash1;
    my %hash2;
    my %hash;
    # Matice i×j ... rozsahy složek; uvnitř label 1 a label 2.
    my @matice;
    foreach my $slozka (@{$strom1->{slozky}})
    {
        my $klic = "$slozka->{i} $slozka->{j} $slozka->{label}";
        $hash1{$klic}++;
        $hash{$klic} = $slozka;
        push(@{$matice[$slozka->{i}][$slozka->{j}][0]}, $slozka->{label});
    }
    foreach my $slozka (@{$strom2->{slozky}})
    {
        my $klic = "$slozka->{i} $slozka->{j} $slozka->{label}";
        $hash2{$klic}++;
        $hash{$klic} = $slozka;
        push(@{$matice[$slozka->{i}][$slozka->{j}][1]}, $slozka->{label});
    }
    # Projít rozsahy a hledat rozdíly v označení složek.
    for(my $i = 0; $i<=$#matice; $i++)
    {
        my $radek = $matice[$i];
        for(my $j = $i; $j<=$#{$radek}; $j++)
        {
            my $bunka = $radek->[$j];
            # Rozsah od i do j.
            # Jestliže je na tomto rozsahu několik vnořených složek, nezajímá nás, která je výše a která hlouběji.
            # Proto je seřadíme abecedně.
            my $slozky1 = join(":", sort(@{$bunka->[0]}));
            my $slozky2 = join(":", sort(@{$bunka->[1]}));
            # Zajímají nás případy, kdy se zpracování daného rozsahu v obou stromech liší.
            if($slozky1 ne $slozky2)
            {
                zaznamenat_chybu($i, $j, $slozky1, $slozky2);
            }
        }
    }
#    print("\n");
}



#------------------------------------------------------------------------------
# Porovná složky ve třech stromech téže věty. První považuje za vzorový rozbor,
# další dva za výstupy dvou různých parserů. Eviduje pouze chyby, které udělal
# druhý parser a první ne.
#------------------------------------------------------------------------------
sub porovnat_3_stromy
{
    # Strom je hash s odkazy na @slova a @slozky.
    my @stromy = @_;
    # Matice i×j ... rozsahy složek; uvnitř názvy složek na daném rozsahu podle jednotlivých stromů
    my @matice;
    for(my $k = 0; $k<=$#stromy; $k++)
    {
        my $strom = $stromy[$k];
        foreach my $slozka (@{$strom->{slozky}})
        {
            push(@{$matice[$slozka->{i}][$slozka->{j}][$k]}, $slozka->{label});
        }
    }
    # Projít rozsahy a hledat rozdíly v označení složek.
    local @chybne_znacky;
    for(my $d = 0; $d<=$#matice; $d++)
    {
        for(my $i = 0; $i+$d<=$#matice; $i++)
        {
            my $j = $i+$d;
            my $bunka = $matice[$i][$j];
            # Rozsah od i do j.
            # Jestliže je na tomto rozsahu několik vnořených složek, nezajímá nás, která je výše a která hlouběji.
            # Proto je seřadíme abecedně.
            my @slozky;
            for(my $k = 0; $k<=$#{$bunka}; $k++)
            {
                $slozky[$k] = join(":", sort(@{$bunka->[$k]}));
            }
            # Zajímají nás případy, kdy první parser neudělal chybu a druhý ano.
            if($slozky[1] eq $slozky[0] && $slozky[2] ne $slozky[0])
            {
                zaznamenat_chybu($i, $j, $slozky[0], $slozky[2]);
            }
        }
    }
}



#------------------------------------------------------------------------------
# Zaznamená všechny možné statistické údaje o chybě.
#------------------------------------------------------------------------------
sub zaznamenat_chybu
{
    my $i = shift; # Počátek rozsahu chybné složky.
    my $j = shift; # Konec rozsahu chybné složky.
    my $slozka1 = shift; # Složka v levém stromu.
    my $slozka2 = shift; # Složka v pravém stromu.
    my $delka = $j-$i;
    # Pro rozsahy větší než 1 zjistit, jestli obsahují alespoň jednu chybnou značku.
    my $contag;
    if($delka>1)
    {
        for(my $k = $i; $k<$j; $k++)
        {
            if($chybne_znacky[$k])
            {
                $contag = " contag";
                last;
            }
        }
    }
    # V obou stromech je na daném rozsahu alespoň jedna složka, ale liší se označení.
    if($slozka1 ne "" && $slozka2 ne "")
    {
#                    printf("%3d %3d label %s %s\n", $i, $j, $slozka1, $slozka2);
        if($delka==1)
        {
            $chyby{"tag $slozka1 $slozka2"}++;
            $chyby{"tag $slozka1 *"}++;
            $chyby{"tag * $slozka2"}++;
            $chyby{"tag"}++;
            $chybne_znacky[$i] = 1;
        }
        else
        {
            $chyby{"$delka label $slozka1 $slozka2"}++;
            $chyby{"label $slozka1 $slozka2"}++;
            $chyby{"label $slozka1 *"}++;
            $chyby{"label * $slozka2"}++;
            $chyby{"label$contag"}++;
            $chyby{"span $delka"}++;
        }
    }
    # V tomto rozsahu je složka pouze v levém stromu.
    elsif($slozka1 ne "")
    {
#                    printf("%3d %3d left  %s\n", $i, $j, $slozka1);
        $chyby{"$delka left $slozka1"}++;
        $chyby{"left $slozka1"}++;
        $chyby{"left$contag"}++;
    }
    # V tomto rozsahu je složka pouze v pravém stromu.
    else
    {
#                    printf("%3d %3d right %s\n", $i, $j, $slozka2);
        $chyby{"$delka right $slozka2"}++;
        $chyby{"right $slozka2"}++;
        $chyby{"right$contag"}++;
    }
    $n_chyb++;
}



#------------------------------------------------------------------------------
# Vypíše rozebranou větu.
#------------------------------------------------------------------------------
sub vypsat_vetu
{
    my $veta = shift; # hash: slova, slozky
    my @slova = @{$veta->{slova}};
    my @slozky = @{$veta->{slozky}};
    # Vypsat slova.
    for(my $i = 0; $i<=$#slova; $i++)
    {
        if($i>0)
        {
            print(" ");
        }
        print("$i $slova[$i]");
    }
    print(" ", $#slova+1, "\n");
    # Vypsat složky.
    foreach my $slozka (@slozky)
    {
        my $i = $slozka->{i};
        my $j = $slozka->{j};
        printf("\t%3d %3d %s\t%s\n", $i, $j, $slozka->{label}, join(" ", @slova[$i..($j-1)]));
    }
    print("\n");
}
