#!/usr/bin/perl
# Porovná dvě obecné statistiky (frekvenční seznamy).
# Formát řádku: 3 pole oddělená tabulátory: událost, absolutní četnost, relativní četnost.
# (c) 2007 Dan Zeman <zeman@ufal.mff.cuni.cz>
# Licence: GNU GPL

sub usage
{
    print STDERR ("Usage: cmp_stat.pl stat1.txt stat2.txt > comparison.txt\n");
}

use utf8;
use open ":utf8";
binmode(STDIN, ":utf8");
binmode(STDOUT, ":utf8");
binmode(STDERR, ":utf8");

# Zkontrolovat počet parametrů.
if(scalar(@ARGV) != 2)
{
    usage();
    die;
}
$hash1 = nacist_statistiku($ARGV[0]);
$hash2 = nacist_statistiku($ARGV[1]);
# Pro každou značku zjistit rozdíl relativních četností.
foreach my $znacka (keys(%{$hash1}))
{
    $rozdily{$znacka} = abs($hash1->{$znacka}{rc}-$hash2->{$znacka}{rc});
}
foreach my $znacka (keys(%{$hash2}))
{
    unless(exists($rozdily{$znacka}))
    {
        $rozdily{$znacka} = abs($hash1->{$znacka}{rc}-$hash2->{$znacka}{rc});
    }
}
# Seřadit značky sestupně podle rozdílu v relativní četnosti.
@seraz = sort {$rozdily{$b}<=>$rozdily{$a}} (keys(%rozdily));
# Zjistit celkovou vzdálenost obou statistik.
foreach my $znacka (@seraz)
{
    $rozdil += $rozdily{$znacka};
}
$rozdil /= 2;
print("celková vzdálenost = $rozdil\n");
# Vypsat přehled značek seřazený podle rozdílu.
# Sloupce: značka (událost); rozdíl rel. č.; levá rel. č.; pravá rel. č.; rozdíl přepoč. na levou abs. č.; totéž na pravou; levá abs; pravá abs
print("událost\tvzdálenost\tlrel\tprel\tlvzd\tpvzd\tlabs\tpabs\n");
foreach my $znacka (@seraz)
{
    my $vzd = $rozdily{$znacka};
    my $lrel = $hash1->{$znacka}{rc};
    my $prel = $hash2->{$znacka}{rc};
    my $labs = $hash1->{$znacka}{ac};
    my $pabs = $hash2->{$znacka}{ac};
    my $lvzd = $lrel>0 ? $vzd*$labs/$lrel : 0;
    my $pvzd = $prel>0 ? $vzd*$pabs/$prel : 0;
    printf("%s\t%f\t%f\t%f\t%d\t%d\t%d\t%d\n", $znacka, $vzd, $lrel, $prel, $lvzd+0.5, $pvzd+0.5, $labs, $pabs);
}



###############################################################################
# PODPROGRAMY
###############################################################################



#------------------------------------------------------------------------------
# Načte výstup z csts_tag_statistics.pl.
#------------------------------------------------------------------------------
sub nacist_statistiku
{
    my $soubor = shift;
    my %hash;
    open(SOUBOR, $soubor) or die("Nelze číst z $soubor: $!\n");
    while(<SOUBOR>)
    {
        # Odstranit konec řádku.
        s/\r?\n$//;
        # Rozdělit řádek na jednotlivé hodnoty.
        my ($znacka, $acetnost, $rcetnost) = split(/\t/, $_);
        # Přeskočit případné prázdné řádky.
        next if($znacka =~ m/^\s*$/);
        # Uložit hodnoty pod názvem značky.
        if(exists($hash{$znacka}))
        {
            die("Duplicitní statistika pro značku '$znacka'.\n");
        }
        $hash{$znacka} = {"ac" => $acetnost, "rc" => $rcetnost};
    }
    close(SOUBOR);
    return \%hash;
}
