#!/usr/bin/perl
# Porovná dva výstupy z csts_tag_statistics.pl.
# (c) 2007 Dan Zeman <zeman@ufal.mff.cuni.cz>
# Licence: GNU GPL

sub usage
{
    print STDERR ("Usage: cmp_tagstat.pl tagstat1.txt tagstat2.txt > comparison.txt\n");
}

use utf8;
use open ":utf8";
binmode(STDIN, ":utf8");
binmode(STDOUT, ":utf8");
binmode(STDERR, ":utf8");
use tagset::sv::hajic;
use tagset::cs::pdt;

# Zkontrolovat počet parametrů.
if(scalar(@ARGV) != 2)
{
    usage();
    die;
}
$hash1 = nacist_statistiku($ARGV[0]);
$hash2 = nacist_statistiku($ARGV[1]);
# Pro každou značku zjistit rozdíl relativních četností.
foreach my $znacka (keys(%{$hash1}))
{
    $rozdily{$znacka} = abs($hash1->{$znacka}{rc}-$hash2->{$znacka}{rc});
}
foreach my $znacka (keys(%{$hash2}))
{
    unless(exists($rozdily{$znacka}))
    {
        $rozdily{$znacka} = abs($hash1->{$znacka}{rc}-$hash2->{$znacka}{rc});
    }
}
# Seřadit značky sestupně podle rozdílu v relativní četnosti.
@seraz = sort {$rozdily{$b}<=>$rozdily{$a}} (keys(%rozdily));
# Vypsat přehled značek seřazený podle rozdílu.
foreach my $znacka (@seraz)
{
    my $features = tagset::sv::hajic::decode($znacka);
    my $pdt = tagset::cs::pdt::encode($features);
    printf("%s\t%f\t%f\t%f\t%d\t%d\t%s\n", $znacka, $rozdily{$znacka}, $hash1->{$znacka}{rc}, $hash2->{$znacka}{rc}, $hash1->{$znacka}{ac}, $hash2->{$znacka}{ac}, $pdt);
    $rozdil += $rozdily{$znacka};
}
print("celkový rozdíl = $rozdil\n");



###############################################################################
# PODPROGRAMY
###############################################################################



#------------------------------------------------------------------------------
# Načte výstup z csts_tag_statistics.pl.
#------------------------------------------------------------------------------
sub nacist_statistiku
{
    my $soubor = shift;
    my %hash;
    open(SOUBOR, $soubor) or die("Nelze číst z $soubor: $!\n");
    while(<SOUBOR>)
    {
        # Odstranit konec řádku.
        s/\r?\n$//;
        # Rozdělit řádek na jednotlivé hodnoty.
        my ($znacka, $acetnost, $rcetnost) = split(/\t/, $_);
        # Přeskočit případné prázdné řádky.
        next if($znacka =~ m/^\s*$/);
        # Uložit hodnoty pod názvem značky.
        if(exists($hash{$znacka}))
        {
            die("Duplicitní statistika pro značku '$znacka'.\n");
        }
        $hash{$znacka} = {"ac" => $acetnost, "rc" => $rcetnost};
    }
    close(SOUBOR);
    return \%hash;
}
