#!/usr/bin/perl
# Converts corpus from CoNLL-06 format to plain tokenized text (1 sentence per line).
# (c) 2007 Dan Zeman <zeman@ufal.mff.cuni.cz>
# License: GNU GPL

use utf8;
use open ":utf8";
binmode(STDIN, ":utf8");
binmode(STDOUT, ":utf8");

$i_sentence = 1;
$new_sentence = 1;
while(<>)
{
    # Blank line signals new sentence.
    # Take several consecutive blank lines as one sentence break (avoid empty sentences).
    if(!$new_sentence && m/^\s*$/)
    {
        print(join(" ", @sentence), "\n");
        splice(@sentence);
        $i_sentence++;
        $new_sentence = 1;
    }
    # Any non-blank line is a token.
    else
    {
        if($new_sentence)
        {
            $new_sentence = 0;
        }
        # Strip the line break.
        s/\r?\n$//;
        # Get field values and print them.
        # Note: $pdttag is not part of CoNLL format but can be added by a tag conversion utility.
        # It can combine information from $cpostag, $postag, and $feats.
        my ($id, $form, $lemma, $cpostag, $postag, $feats, $head, $deprel, $phead, $pdeprel, $pdttag) = split(/\s+/, $_);
#        $cpostag = $postag;
        $form =~ s/&/&amp;/g;
        $form =~ s-/-&slash;-g;
        $form =~ s/ /&space;/g;
        $cpostag =~ s/&/&amp;/g;
        $cpostag =~ s-/-&slash;-g;
        $cpostag =~ s/ /&space;/g;
        push(@sentence, "$form/$cpostag");
    }
}
