#!/usr/bin/perl
# Delexicalizes CSTS trees, i.e. copies morphological tags over word forms.
# (c) 2006-2007 Dan Zeman <zeman@ufal.mff.cuni.cz>
# License: GNU GPL

sub usage
{
    print STDERR ("Usage: delex.pl [-f tagset0 -t tagset1] < input.csts > output.csts\n");
    print STDERR ("  input:   CSTS file\n");
    print STDERR ("  output:  same file but tags will be copied over word forms\n");
    print STDERR ("  tagset0: copied tags are in this tagset; e.g. 'da::conll'\n");
    print STDERR ("  tagset1: the copy will be converted into this tagset\n");
}

use utf8;
use open ":utf8";
binmode(STDIN, ":utf8");
binmode(STDOUT, ":utf8");
binmode(STDERR, ":utf8");
use csts1;
use Getopt::Long;



# Get options.
GetOptions
(
    'from=s' => \$tagset0,
    'to=s'  => \$tagset1,
);
if($tagset0 ne "" && $tagset1 ne "")
{
    my $eval;
    $eval .= "use tagset::$tagset0;\n";
    $eval .= "use tagset::$tagset1;\n";
    $eval .= '$decode = \&tagset::'.$tagset0.'::decode;'."\n";
    $eval .= '$encode = \&tagset::'.$tagset1.'::encode;'."\n";
    eval($eval);
}
else
{
    $decode = $encode = sub {return $_[0]};
}
csts::cist(\&zpracovat_vetu);



#------------------------------------------------------------------------------
# Processes a tree.
#------------------------------------------------------------------------------
sub zpracovat_vetu
{
    my $veta = shift;
    # Copy tags over word forms in each node.
    foreach my $uzel (@{$veta})
    {
        $uzel->{form} = &{$encode}(&{$decode}($uzel->{tag}));
        # Some parsers do not like whitespaces in words while CoNLL tags may contain tabs.
        $uzel->{form} =~ s/\s+/_/g;
    }
    # Print tree.
    csts::psat_vetu($veta);
}
