#!/usr/bin/perl
# Detokenizes text. Erases spaces before commas, periods etc.
# Of course cannot always restore original text before tokenization.
# (c) 2007 Dan Zeman <zeman@ufal.mff.cuni.cz>
# License: GNU GPL

sub usage
{
    print STDERR ("Usage: detok.pl [-l language] < input.tok > output.txt\n");
    print STDERR ("  language: ISO 639-1 code; helps interpret quotation marks; default = en\n");
}

use utf8;
use open ":utf8";
binmode(STDIN, ":utf8");
binmode(STDOUT, ":utf8");
binmode(STDERR, ":utf8");
use Getopt::Long;

# Get options.
$lang = "en";
GetOptions
(
    'language=s' => \$lang,
);
# Warning! The following constants are language-dependent!
# Quotation marks: English high66-high99, Czech and German low99-high66.
$lqm = $lang eq "en" ? chr(8220) : chr(8222);
$rqm = $lang eq "en" ? chr(8221) : chr(8220);

while(<>)
{
    # Erase spaces after left bracket etc.
    s/([\(\[\{$lqm]|\`\`)\s+/$1/g;
    # Erase spaces before period, comma, semicolon, colon, right bracket etc.
    s/\s+([\.\!\?,;:\)\]\}$rqm]|'')/$1/g;
    # Print the modified line.
    print;
}
