#!/usr/bin/perl
# Searches file system subtree for result.txt files.
# Assumes those are evalb reports.
# Collects selected values from all reports and prints them to STDOUT with paths to the reports.
# (c) 2007 Dan Zeman <zeman@ufal.mff.cuni.cz>
# License: GNU GPL

use utf8;
use open ":utf8";
binmode(STDIN, ":utf8");
binmode(STDOUT, ":utf8");
binmode(STDERR, ":utf8");

search(".", \@results);
foreach my $result (sort {$a->{path} cmp $b->{path}} @results)
{
    printf("P = %6.2f\tR = %6.2f\tF = %6.2f\tT = %6.2f\tN = %3d\t%s\n", $result->{p}, $result->{r}, $result->{f}, $result->{t}, $result->{n}, $result->{path});
}

#------------------------------------------------------------------------------
# Searches file system subtree. Whenever it finds a result.txt, it reads the
# file (assuming it is an evalb output) and adds its values to the global list.
#------------------------------------------------------------------------------
sub search
{
    my $path = shift; # path to the folder (subtree root) to search
    my $list = shift; # reference to array of collected hashes of results
    # If there is a result.txt, read it.
    my $resultpath = $path."/result.txt";
    if(-f $resultpath)
    {
        my $hash = read_report($resultpath);
        $hash->{path} = $path;
        push(@{$list}, $hash);
    }
    # Search through subfolders.
    opendir(DIR, $path) or die("Cannot read folder $path: $!\n");
    my @files = readdir(DIR);
    closedir(DIR);
    foreach my $file (@files)
    {
        my $subfpath = $path."/".$file;
        if($file !~ m/^\.\.?$/ && -d $subfpath)
        {
            search($subfpath, $list);
        }
    }
}

#------------------------------------------------------------------------------
# Reads a report made by evalb. Returns reference to hash with selected values.
#------------------------------------------------------------------------------
sub read_report
{
    my $path = shift; # path to the file with the report
    my %hash;
    open(REPORT, $path) or die("Cannot read report $path: $!\n");
    while(<REPORT>)
    {
        # Remove line break.
        s/\r?\n$//;
        # Most values occur twice: for all sentences and for len<=40.
        # Later values override the former, which means we prefer the values for len<=40.
        if(m/Bracketing\s+Precision\s+=\s+(\d+\.\d+)/i)
        {
            $hash{p} = $1;
        }
        elsif(m/Bracketing\s+Recall\s+=\s+(\d+\.\d+)/i)
        {
            $hash{r} = $1;
        }
        elsif(m/Bracketing\s+FMeasure\s+=\s+(\d+\.\d+)/i)
        {
            $hash{f} = $1;
        }
        elsif(m/Tagging\s+accuracy\s+=\s+(\d+\.\d+)/i)
        {
            $hash{t} = $1;
        }
        elsif(m/Number\s+of\s+Valid\s+sentence\s+=\s+(\d+)/i)
        {
            $hash{n} = $1;
        }
    }
    close(REPORT);
    return \%hash;
}
