#!/usr/bin/perl
# Naivní dělení odstavců na věty v cizojazyčném textu.
# Vstupní text je tokenizovaný a na každém řádku je jeden odstavec.
# Na výstupu je na každém řádku jedna věta.
# (c) 2006 Dan Zeman <zeman@ufal.mff.cuni.cz>
# Licence: GNU GPL

use utf8;
use open ":utf8";
binmode(STDIN, ":utf8");
binmode(STDOUT, ":utf8");
binmode(STDERR, ":utf8");

while(<>)
{
    # Odstranit konec řádku a přebytečné mezery na začátku a na konci.
    s/\r?\n$//;
    s/^\s+//;
    s/\s+$//;
    # Převést odstavec na pole slov.
    my @slova = split(/\s+/, $_);
    # Projít slova a hledat tečky (neměly by být přilepené ke slovu, protože text je tokenizovaný).
    for(my $i = 1; $i<$#slova; $i++)
    {
        # Koncem věty může být posloupnost teček (typicky jedna nebo tři), vykřičník nebo otazník. Zatím ignorujeme závorky a uvozovky.
        next unless($slova[$i] =~ m/^(\.+|!|\?)$/);
        # Následující slovo musí začínat velkým písmenem. (Ale pozor, neznáme seznam písmen v abecedě daného jazyka!)
        # Následující slovo také mohou být uvozovky (neumíme říci, zda patří k větě vlevo, nebo vpravo, necháme je vždy vpravo).
        # Následující slovo také může být pomlčka (patří větě vpravo).
        # Následující slovo také může být pravá závorka (patří větě vlevo) nebo levá závorka (patří větě vpravo).
        next unless($slova[$i+1] =~ m/^\p{upper}/ || $slova[$i+1] =~ m/^[-"\(\)]$/);
        $i++ if($slova[$i+1] eq ")");
        # Předcházející slovo nesmí být číslo menší než 1000. Výjimka: před číslem je znak dolaru.
        next if($slova[$i-1] =~ m/\d$/ && $slova[$i-1]<1000 && !($i>1 && $slova[$i-2] eq '$'));
        # Předcházející slovo nesmí být jediné velké písmeno.
        next if($slova[$i-1] =~ m/^\p{upper}$/);
        # Předcházející slovo nesmí být na seznamu častých zkratek.
        next if($slova[$i-1] =~ m/^(nr|ii|iii|iv|vi|vii|viii|ix|xi|xii|xiii|xiv|xv|xvi|xvii|xviii|xix|xx|th|ph|dr|hr)$/i);
        # Pokud jsme došli až sem, asi je to opravdu konec věty.
        $slova[$i] .= "\n";
    }
    # Znova slepit původní odstavec.
    my $odstavec = join(" ", @slova)."\n";
    # Umazat mezery na začátku řádku (vznikly na začátku všech vět kromě té první).
    $odstavec =~ s/\n\s+/\n/sg;
    # Poslat odstavec na výstup.
    print($odstavec);
}
