#!/usr/bin/perl
# Repairs tokenization in parser output to ensure all sentences can be evaluated.
# (c) 2006 Dan Zeman <zeman@ufal.mff.cuni.cz>
# License: GNU GPL

use utf8;
use open ":utf8";
binmode(STDIN, ":utf8");
binmode(STDOUT, ":utf8");
binmode(STDERR, ":utf8");

unless(scalar(@ARGV)>=1)
{
    print STDERR ("Usage: goldretok.pl tokenizedTextParserInput bracketedParserOutput > refinedParserOutput\n");
    print STDERR ("Or:    goldretok.pl tokenizedTextParserInput < bracketedParserOutput > refinedParserOutput\n");
    die("Missing argument 1: tokenized plain text parser input.\n");
}
# Open plain text (desired tokenization) for reading.
$desired = shift(@ARGV); # shift to enable while(<>) later
open(DESIRED, $desired) or die("Cannot read from $desired: $!\n");
# Read parser output. One tree must be on one line!
while(<>)
{
    # Skip empty lines.
    next if(m/^\s*$/);
    # Get rid of the line break.
    s/\r?\n$//;
    # Insert spaces before right rounded brackets.
    s/\)/ \)/g;
    # Split tree into words and brackets (nonterminal are stuck to left rounded brackets).
    @tokens = split(/\s+/, $_);
    # Read the corresponding line in the desired tokenization.
    $desline = <DESIRED>;
    # Get rid of the line break.
    $desline =~ s/\r?\n$//;
    # Split it into tokens.
    @destokens = split(/\s+/, $desline);
    # Loop through tree tokens. Leave brackets and nonterminals; replace terminals by desired tokens.
    ###!!! We are currently ignoring that a token could be split and thus sentence lenght would not match!
    for(my $i = 0; $i<=$#tokens; $i++)
    {
        # Skip bracketing.
        next if($tokens[$i] =~ m/^[\(\)]/);
        # It is a terminal token. Now get the next desired token. But skip traces and similar junk!
        my $destok = shift(@destokens);
        while($destok =~ m/^\*/ || $destok eq "0")
        {
            $destok = shift(@destokens);
        }
        # If anything went wrong and we reached the end of @destokens, we have an empty token.
        # It is probably safer to replace it with something non-empty.
        if($destok =~ m/^\s*$/)
        {
            $destok = "_EMPTY_";
        }
        # Compare actual token with the desired one.
        # Report any discrepancies.
        # But make sure the desired token will be on the output.
        if($tokens[$i] ne $destok)
        {
            print STDERR ("actual = $tokens[$i]\tdesired = $destok\n");
        }
        $tokens[$i] = $destok;
    }
    # Join tokens back to one line and print it.
    print(join(" ", @tokens), "\n");
}
close(DESIRED);
