#!/usr/bin/perl
# Converts North American News Text SGML to plain text.
# Puts one paragraph on one line.
# (c) 2006 Dan Zeman <zeman@ufal.mff.cuni.cz>
# Licence: GNU GPL

use utf8;
use open ":utf8";
binmode(STDIN, ":utf8");
binmode(STDOUT, ":utf8");
binmode(STDERR, ":utf8");
# We cannot use an XML parser because the data does not contain </p> end tags.
use HTML::Parser();

# Check that we have the argument.
unless(scalar(@ARGV)==1)
{
    print STDERR ("Usage: nant2txt.pl nant.sgml\n");
    print STDERR ("Wrong number of arguments.\n");
    die();
}
# Create HTML parser.
$p = HTML::Parser->new
(
    api_version => 3,
    start_h => [\&start, "tagname, attr"],
    text_h  => [\&text, "dtext"],
    end_h   => [\&end, "tagname"],
    marked_sections => 1
);
# Parse the SGML file.
$p->parse_file($ARGV[0]);



###############################################################################
# SUBROUTINES
###############################################################################



#------------------------------------------------------------------------------
# Handles SGML start tags.
#------------------------------------------------------------------------------
sub start
{
    my $tagname = shift;
    my $attr = shift; # reference to hash
    # Remember if we are inside of <text>...</text>.
    if($tagname eq "text")
    {
        $text = 1;
    }
    # Start of new paragraph means end of the preceding one.
    elsif($tagname eq "p")
    {
        # If this is not the first paragraph and if the preceding paragraph
        # is not empty, print the previous paragraph.
        unless($par =~ m/^\s*$/)
        {
            # Remove excess spaces including newlines.
            $par =~ s/^\s+//s;
            $par =~ s/\s+$//s;
            $par =~ s/\s+/ /gs;
            # Print paragraph to standard output.
            print("$par\n");
            # Clear the variable for the next paragraph.
            $par = "";
        }
    }
}



#------------------------------------------------------------------------------
# Handles text between tags.
#------------------------------------------------------------------------------
sub text
{
    my $dtext = shift; # text with entities decoded
    $par .= $dtext;
}



#------------------------------------------------------------------------------
# Handles SGML end tags.
#------------------------------------------------------------------------------
sub end
{
    my $tagname = shift;
    # Remember if we are inside of <text>...</text>.
    if($tagname eq "text")
    {
        $text = 0;
    }
}
