#!/usr/bin/perl
# Projde soubor složkových stromů, spočítá jednotlivé druhy složek a vypíše statistiku.
# (c) 2007 Dan Zeman <zeman@ufal.mff.cuni.cz>
# Licence: GNU GPL

sub usage
{
    print STDERR ("Usage: penn_phr_statistics.pl < input.penn > report.txt\n");
}

use utf8;
use open ":utf8";
binmode(STDIN, ":utf8");
binmode(STDOUT, ":utf8");
binmode(STDERR, ":utf8");
use penn;



$n = 0;
while(1)
{
    $text = nacist_vetu(STDIN);
    last if($text eq "");
    $strom = penn::decompose($text);
    # Upravit statistiku složek.
    foreach my $slozka (@{$strom->{slozky}})
    {
        my $pravidlo = "$slozka->{label} --> $slozka->{rhs}";
        $gramatika{$pravidlo}++;
        $n++;
    }
}
# Vypsat gramatiku.
@klice = sort {$gramatika{$b} <=> $gramatika{$a}} keys(%gramatika);
foreach my $pravidlo (@klice)
{
    $r = $n>0 ? $gramatika{$pravidlo}/$n : 0;
    print("$pravidlo\t$gramatika{$pravidlo}\t$r\n");
}



#------------------------------------------------------------------------------
# Načte větu z otevřeného souboru. Počítá s tím, že jedna věta může být na více
# než jednom řádku. Předpokládá ale, že na jednom řádku nejsou části více než
# jedné věty.
#------------------------------------------------------------------------------
sub nacist_vetu
{
    my $SOUBOR = shift;
    # Vždy, když se vyrovná počet levých a pravých závorek, končí věta.
    # Řetězce přilepené k levým závorkám jsou neterminály.
    # Ostatní řetězce jsou vesměs slova (tokeny), ale "*-1", "*-2"..., "*U*", "*" je stopa nebo něco takového.
    my $text;
    my $nlz_celkem = 0;
    my $npz_celkem = 0;
    while(<$SOUBOR>)
    {
        s/\r?\n$//;
        # Spočítat levé a pravé závorky v právě načteném řádku.
        my ($nlz, $npz);
        my $lz = $_;
        $lz =~ s/[^(]//g;
        $nlz = length($lz);
        my $pz = $_;
        $pz =~ s/[^)]//g;
        $npz = length($pz);
        # Přičíst počty závorek k celkovým počtům a zjistit, zda jde o konec věty.
        $nlz_celkem += $nlz;
        $npz_celkem += $npz;
        my $konec_vety = $nlz_celkem>0 && $nlz_celkem==$npz_celkem;
        # Připojit aktuální řádek k větě, a pokud je věta celá, přerušit čtení a zpracovat ji.
        $text .= $_;
        last if($konec_vety);
    }
    # Jestliže text neobsahuje ani jednu závorku, nenačetli jsme větu (strom), ale nanejvýš prázdné řádky mezi.
    # Možná jsme také na konci souboru. Usnadníme detekci takových případů a text předáme rovnou zcela prázdný.
    if($text !~ m/[\(\)]/)
    {
        $text = "";
    }
    return $text;
}



#------------------------------------------------------------------------------
# Vypíše rozebranou větu.
#------------------------------------------------------------------------------
sub vypsat_vetu
{
    my $veta = shift; # hash: slova, slozky
    my @slova = @{$veta->{slova}};
    my @slozky = @{$veta->{slozky}};
    # Vypsat slova.
    for(my $i = 0; $i<=$#slova; $i++)
    {
        if($i>0)
        {
            print(" ");
        }
        print("$i $slova[$i]");
    }
    print(" ", $#slova+1, "\n");
    # Vypsat složky.
    foreach my $slozka (@slozky)
    {
        my $i = $slozka->{i};
        my $j = $slozka->{j};
        printf("\t%3d %3d %s\t%s\n", $i, $j, $slozka->{label}, join(" ", @slova[$i..($j-1)]));
    }
    print("\n");
}
