#!/usr/bin/perl
# Rozdělí CoNLL soubor na prvních N vět a zbytek.
# Copyright © 2007, 2009 Dan Zeman <zeman@ufal.mff.cuni.cz>
# Licence: GNU GPL
# 29.3.2009: přidána volba -phead (procenta místo absolutního počtu vět)

sub usage
{
    print STDERR ("Usage: split_conll.pl < input [-head N] [-phead N] output1 output2\n");
    print STDERR ("  head:    how many initial sentences go to output1\n");
    print STDERR ("  phead:   same as head but argument is percentage of the whole input\n");
    print STDERR ("  output1: path where to put the first HEAD sentences\n");
    print STDERR ("  output2: path where to put the rest\n");
}

use utf8;
use open ":utf8";
binmode(STDIN, ":utf8");
binmode(STDOUT, ":utf8");
binmode(STDERR, ":utf8");
use Getopt::Long;



# Get options.
$head = 0;
$phead = 0;
GetOptions('head=i' => \$head, 'phead=i' => \$phead);
if(scalar(@ARGV)<2)
{
    usage();
    die;
}
# Odebrat cesty k výstupům z @ARGV, aby je tam neviděla čtecí procedura a nepovažovala je za vstupy.
$vystup[0] = shift(@ARGV);
$vystup[1] = shift(@ARGV);
# Pokud máme použít -phead, musíme nejdřív proběhnout vstup a spočítat věty.
# Situaci nám komplikuje, že čteme ze standardního vstupu, který nemůžeme otevřít znova a čekat v něm to samé.
if($phead && !$head)
{
    my $n_sentences = 0;
    $tmp = "/tmp/$$";
    open(TMP, ">$tmp") or die("Cannot write to temporary file $tmp: $!\n");
    while(<>)
    {
        # Prázdný řádek signalizuje novou větu.
        # Předpokládáme, že i za poslední větou na vstupu je právě jeden prázdný řádek.
        # Také předpokládáme (na rozdíl od staršího kódu níže), že neexistují prázdné věty (více prázdných řádků za sebou).
        if(m/^\s*$/)
        {
            $n_sentences++;
        }
        # Uložit si vstupní řádek do pomocného souboru.
        print TMP;
    }
    close(TMP);
    # Vypočítat $head na základě $phead a $n_sentences.
    $head = int($n_sentences*$phead/100);
    # Podstrčit pomocný soubor, kam jsme zkopírovali vstup, jako falešný argument, kterého se chytne <>.
    splice(@ARGV);
    $ARGV[0] = $tmp;
}
# Otevřít výstup 1.
open(OUTPUT, ">$vystup[0]") or die("Cannot write to $vystup[0]: $!\n");
$i_sentence = 1;
$new_sentence = 1;
while(<>)
{
    # Blank line signals new sentence.
    if(m/^\s*$/)
    {
        # Take several consecutive blank lines as one sentence break (avoid empty sentences).
        if(!$new_sentence)
        {
            $new_sentence = 1;
            $i_sentence++;
            # Jestliže je to první věta druhého souboru, zavřít výstup 1 a otevřít výstup 2.
            if($i_sentence==$head+1)
            {
                close(OUTPUT);
                open(OUTPUT, ">$vystup[1]") or die("Cannot write to $vystup[1]: $!\n");
            }
            else
            {
                print OUTPUT ("\n");
            }
        }
    }
    # Any non-blank line is a token.
    else
    {
        $new_sentence = 0;
        print OUTPUT;
    }
}
close(OUTPUT);
# Uklidit po sobě pomocný soubor.
if($tmp)
{
    unlink($tmp);
}
