#!/usr/bin/perl
# Tags Swedish CSTS corpus using Hajič Swedish tagger.
# (c) 2007 Dan Zeman <zeman@ufal.mff.cuni.cz>
# License: GNU GPL

use utf8;
binmode(STDIN, ":utf8");
binmode(STDOUT, ":utf8");
binmode(STDERR, ":utf8");
use Encode;

# Create temporary working folder.
$workdir = "/tmp/hajicsv.$$";
mkdir($workdir) or die("Cannot create folder $workdir: $!\n");
# Convert input text from UTF-8 to ISO 8859-1. Hajič's tagger cannot handle UTF-8.
open(PREPROC, ">$workdir/input.iso.csts") or die("Cannot write to file $workdir/input.iso.csts: $!\n");
while(<>)
{
    # Strip previous morphological annotation, if any.
    s/<l>[^<]*<t>[^<]*//;
    print PREPROC (encode("iso-8859-1", $_));
}
close(PREPROC);
# Run Hajič's tagger.
$tagger = "/home/zeman/nastroje/taggery/hajic-sv/2006-11-08/SE061108x";
system("$tagger TG $workdir/input.iso.csts $workdir/output.iso.csts");
# Convert output's encoding and send it to STDOUT.
open(POSTPROC, "$workdir/output.iso.csts") or die("Cannot read from file $workdir/output.iso.csts: $!\n");
while(<POSTPROC>)
{
    # Convert back from ISO 8859-1 to UTF-8.
    $_ = decode("iso-8859-1", $_);
    # Remove ambiguous morphological analysis.
    s/<MM[lt][^>]*>[^<]*//g;
    # Change disambiguated tags markup from <MDt> to <t>.
    s/<MD([lt])[^>]*>/<$1>/g;
    # Correct tags of sentence-final punctuation. For some reason, Hajič always tags it as noun.
    s/(<f>\.+<l>\.+<t>)N[^<]+/$1FE-------/;
    # Send the modified line to STDOUT.
    print;
}
close(POSTPROC);
# Clean up working folder.
system("rm -rf $workdir");
