#!/usr/bin/perl
# Converts tags in CSTS trees from one tagset to another.
# (c) 2006-2007 Dan Zeman <zeman@ufal.mff.cuni.cz>
# License: GNU GPL

sub usage
{
    print STDERR ("Usage: tagconv.pl -f tagset0 -t tagset1 < input.csts > output.csts\n");
    print STDERR ("  input:   CSTS file\n");
    print STDERR ("  output:  same file but tags will be copied over word forms\n");
    print STDERR ("  tagset0: copied tags are in this tagset; e.g. 'da::conll'\n");
    print STDERR ("  tagset1: the copy will be converted into this tagset\n");
}

use utf8;
use open ":utf8";
binmode(STDIN, ":utf8");
binmode(STDOUT, ":utf8");
binmode(STDERR, ":utf8");
use csts1;
use Getopt::Long;



# Get options.
GetOptions
(
    'from=s' => \$tagset0,
    'to=s'  => \$tagset1,
);
if($tagset0 eq "" || $tagset1 eq "")
{
    usage();
    die();
}
else
{
    my $eval;
    $eval .= "use tagset::$tagset0;\n";
    $eval .= "use tagset::$tagset1;\n";
    $eval .= '$decode = \&tagset::'.$tagset0.'::decode;'."\n";
    $eval .= '$encode = \&tagset::'.$tagset1.'::encode;'."\n";
    eval($eval);
}
csts::cist(\&zpracovat_vetu);



#------------------------------------------------------------------------------
# Processes a tree.
#------------------------------------------------------------------------------
sub zpracovat_vetu
{
    my $veta = shift;
    # Copy tags over word forms in each node.
    foreach my $uzel (@{$veta})
    {
        $uzel->{tag} = &{$encode}(&{$decode}($uzel->{tag}));
    }
    # Print tree.
    csts::psat_vetu($veta);
}
