#!/usr/bin/perl
# Extracts two line-by-line aligned plain texts from JRC Acquis alignment file with texts (getAlignmentWithText.pl).
# (c) 2006 Dan Zeman <zeman@ufal.mff.cuni.cz>
# License: GNU GPL

sub usage
{
    print STDERR ("Usage: teialign2txt.pl < input.xml first.txt second.txt\n");
}

use utf8;
use open ":utf8";
binmode(STDIN, ":utf8");
binmode(STDOUT, ":utf8");
binmode(STDERR, ":utf8");
use XML::Parser;

if(scalar(@ARGV)<2)
{
    usage();
    die("Missing argument(s) - output file name(s).\n");
}
$s1 = shift(@ARGV);
$s2 = shift(@ARGV);
$parser = new XML::Parser(Handlers =>
{
    Start => \&handle_start,
    End   => \&handle_end,
    Char  => \&handle_char
});
while(<>)
{
    $xml .= $_;
}
open(S1, ">$s1") or die("Cannot write to $s1: $!\n");
open(S2, ">$s2") or die("Cannot write to $s2: $!\n");
$parser->parse($xml, ProtocolEncoding => 'utf-8', ErrorContext => 1);
close(S1);
close(S2);



###############################################################################
# SUBROUTINES
###############################################################################



#------------------------------------------------------------------------------
# Handles the occurrence of a XML start tag.
#------------------------------------------------------------------------------
sub handle_start
{
    my $expat = shift;
    my $element = shift;
    my %attr = @_;
    if($element eq "teiHeader")
    {
        $header = 1;
    }
    elsif(!$header && $element eq "link")
    {
        $inside = 1;
    }
    # What segment are we reading now?
    elsif($element eq "s1")
    {
        $segment = 1;
    }
    elsif($element eq "s2")
    {
        $segment = 2;
    }
}



#------------------------------------------------------------------------------
# Handles plain text.
#------------------------------------------------------------------------------
sub handle_char
{
    my $expat = shift;
    my $string = shift;
    if($inside)
    {
        $text[$segment] .= $string;
    }
}



#------------------------------------------------------------------------------
# Handles the occurrence of a XML end tag.
#------------------------------------------------------------------------------
sub handle_end
{
    my $expat = shift;
    my $element = shift;
    if($element eq "teiHeader")
    {
        $header = 0;
    }
    elsif($element eq "link")
    {
        # Process both segments.
        for(my $i = 1; $i<=2; $i++)
        {
            # Remove superfluous spaces.
            $text[$i] =~ s/^\s+//s;
            $text[$i] =~ s/\s+$//s;
            $text[$i] =~ s/\s+/ /sg;
        }
        # Skip pairs where one or both segments are empty.
        if($text[1] !~ m/^\s*$/ && $text[2] !~ m/^\s*$/)
        {
            print S1 ("$text[1]\n");
            print S2 ("$text[2]\n");
        }
        $text[0] = "";
        $text[1] = "";
        $text[2] = "";
        $inside = 0;
    }
    elsif($element =~ m/^s[12]$/)
    {
        $segment = "";
    }
    # If segment consists of multiple paragraphs, add space at paragraph end.
    elsif($element eq "p")
    {
        $text[$segment] .= " ";
    }
}
