#!/usr/bin/perl
# Převede tokenizovaný text, ve kterém je na jednom řádku jedna věta, do CSTS.
# (c) 2006 Dan Zeman <zeman@ufal.mff.cuni.cz>
# Licence: GNU GPL

use utf8;
use open ":utf8";
binmode(STDIN, ":utf8");
binmode(STDOUT, ":utf8");
binmode(STDERR, ":utf8");
use Getopt::Long;

# Mandatory information: -l ISO-639 language code.
GetOptions('language=s' => \$lang);
if($lang eq "")
{
    print STDERR ("Usage: tok2csts.pl -l language < input.tok.txt > output.csts\n");
    print STDERR ("Missing language (ISO 639-1 code).\n");
    die;
}
print("<csts lang=\"$lang\">\n");
# omit optional <h> (header)
print("<doc file=\"file\" id=\"1\">\n");
# <a> header of document is not optional
print("<a>\n");
# include only mandatory elements of <a>
print("<mod>s\n"); # s = current language; d = old language; o = transcribed speech
print("<txtype>mix\n"); # ima = imaginative, inf = informative, mix = mixed
print("<genre>mix\n"); # mixed or unknown genre
print("<med>net\n"); # b = book, nws = newspaper, j = journal (periodic), scr = screenplay, net = Internet, oc = occasional
print("<lang>$lang\n"); # language of the document
print("<temp>0000\n"); # four-digit year of publication
print("<opus>file\n"); # according to Czech National Corpus definitions, this shall be equal to filename
print("<id>1\n"); # document identification within a file, equals to the id attribute of <doc>
print("</a>\n");
# All sentences will be in one chapter and one paragraph.
print("<c>\n");
print("<p>\n");
while(<>)
{
    $i_sentence++;
    print("<s id=\"$i_sentence\">\n");
    s/\r?\n$//;
    @slova = split(/\s+/, $_);
    foreach my $slovo (@slova)
    {
        # Escape special characters.
        $slovo =~ s/&/&amp;/g;
        $slovo =~ s/</&lt;/g;
        $slovo =~ s/>/&gt;/g;
        print("<f>$slovo\n");
    }
}
print("</p>\n");
print("</c>\n");
print("</csts>\n");
