#!/usr/bin/perl
# Translates a text word-by-word using a glossary.
# Passes unknown words untranslated.
# (c) 2006 Dan Zeman <zeman@ufal.mff.cuni.cz>
# Licence: GNU GPL

sub usage
{
    print STDERR ("Usage: wbwtranslate.pl -g glossary < input > output\n");
    print STDERR ("  glossary: each line: source word, space(s), target word\n");
    print STDERR ("  input:    STDIN or arguments; must be tokenized\n");
}

use utf8;
use open ":utf8";
binmode(STDIN, ":utf8");
binmode(STDOUT, ":utf8");
binmode(STDERR, ":utf8");
use Getopt::Long;

# Get options.
GetOptions('glossary=s' => \$glossary);
if($glossary eq "")
{
    usage();
    die("Missing glossary file name.\n");
}
# Read glossary.
open(GLOSSARY, $glossary) or die("Cannot read from $glossary: $!\n");
while(<GLOSSARY>)
{
    # Strip the line break.
    s/\r?\n$//;
    # Save the entry.
    my ($source, $target) = split(/\s+/, $_);
    $gloss{$source} = $target;
}
close(GLOSSARY);
# Read standard input, translate it and write it to standard output.
while(<>)
{
    # Strip the line break.
    s/\r?\n$//;
    my @words = split(/\s+/, $_);
    # Translate the words.
    @words = map
    {
        # Classify the source word as either lowercase, uppercase, or capitalized.
        my $contains_lowercase = m/\p{Ll}/;
        my $is_capitalized = m/^\p{Lu}/ && $contains_lowercase;
        my $lowercased = lc($_);
        my $result = exists($gloss{$lowercased}) && $gloss{$lowercased} !~ m/^\s*$/ ? $gloss{$lowercased} : $lowercased;
        # Restore case.
        if($is_capitalized)
        {
            $result =~ s/^(.)/\u$1/;
        }
        elsif(!$contains_lowercase)
        {
            $result = uc($result);
        }
        # Make sure the result is not empty.
        if($result =~ m/^\s*$/)
        {
            die("Empty result for $_!\n");
        }
        $result;
    }
    (@words);
    # Print the translation.
    print(join(" ", @words), "\n");
}
